"""
This module includes classes and methods to store mutations.
Maybe here the randomiation should occur.
"""

import re
import numpy as np
import pandas as pd
from Bio.Seq import Seq
import randommut.refseq as rs



def mutset_from_path(av_input_path, to0base=True):
    """
    Obtain a dictoniary of mutset onbjects by chromosome
    """

    # colum definition
    # 0. chr
    # 1. start
    # 2. end
    # 3. ref
    # 4. alt
    # 5. strand
    # 6. sample

    table = pd.read_csv(av_input_path,
                        header=None,
                        sep="\t",
                        index_col=False)

    chrom_idx = {}
    chrom_unique = np.unique(table.iloc[:, [0]])

    for chrom in chrom_unique:
        idx = np.array(table.iloc[:, [0]] == chrom)
        chrom_idx[chrom] = [i for i, x in enumerate(idx) if x]

    ms_chr = {}
    for chrom in chrom_idx:
        idx = chrom_idx[chrom]
        #tmp_table = table[idx,]
        ms_chr[chrom] = MutSet(chr_id=chrom,
                               pos_start=table.iloc[idx, 1],
                               pos_end=table.iloc[idx, 2],
                               ref=table.iloc[idx, 3],
                               alt=table.iloc[idx, 4],
                               strand=table.iloc[idx, 5],
                               sample_id=table.iloc[idx, 6],
                               to0base=to0base)

    return ms_chr



class MutSet(object):
    " Class to store a set of mutation "
    def __init__(self,
                 chr_id,
                 pos_start,
                 pos_end,
                 sample_id,
                 ref,
                 alt,
                 strand,
                 to0base=True):

        # each mutset have only one chromosome
        if len(chr_id) != 1:
            if len(np.unique(chr_id)) == 1:
                chr_id = str(np.unique(chr_id))
            else:
                raise ValueError

        self.chr_id = chr_id

        # here I check if all lists have the same length
        common_length = len(pos_start)
        if any(len(lst) != common_length for lst in [pos_end,
                                                     sample_id,
                                                     ref,
                                                     alt,
                                                     strand]):
            raise ValueError

        # if input comes from bed file no need for conversion
        if to0base:
            pos_start = np.array(pos_start,dtype=int) - 1
            pos_end = np.array(pos_end,dtype=int)
            self.pos = np.column_stack((pos_start, pos_end))
        else:
            pos_start = np.array(pos_start)
            pos_end = np.array(pos_end)
            self.pos = np.column_stack((pos_start, pos_end))


        strand_re = re.compile("^[-0]$") # could break not tested
        idx = re.finditer(strand_re, str(strand))

        # this is not super fast but most of the time we will have positive
        for i in idx:
            ref[i] = Seq(ref[i]).reverse_complement()
            alt[i] = Seq(alt[i]).reverse_complement()

        self.meta = np.column_stack((sample_id, ref, alt))
    def __len__(self):
        return len(self.pos)
    def get_chr_id(self):
        "return the chromosome id"
        return self.chr_id
    def get_sample(self):
        " return sample as a list "
        return self.meta[:, 0]
    def get_ref(self):
        " return the references as a list "
        return self.meta[:, 1]
    def get_alt(self):
        " return alternative as a list "
        return self.meta[:, 2]
    def get_tr(self):
        " get the transision in MS6 format"
        def make_tr(arr):
            " small function to generate the tr"
            # this uses the A C format (useful to make generalization here)
            if arr[0] in ["T", "G"]:
                arr[0] = Seq(arr[0]).reverse_complement()
                arr[1] = Seq(arr[1]).reverse_complement()
            return "{}>{}".format(arr[0], arr[1])
        res = np.apply_along_axis(make_tr, axis=1, arr=self.meta[:, [1, 2]])
        return res
    def get_context(self, chromosome_object, k=1):
        """
        Return the context for a position in the mask. Needs the 3 masks
        in the correct order (nmask purine and strong). The k indicates
        how many positions we should move to the upstream and downstream.
        1 per default means the 96 context. It uses by default baseset
        C A, meaning that mutations that have a reference value diferent
        than C or A will be transformed to the other strand by reverese
        complementary
        """
        # Important pos_start != pos_end
        left_end = self.pos[:, 0] - k # this is the start
        rigth_end = self.pos[:, 1] + k  # this is the end

        masks = chromosome_object.seq_mask
        ctx = []

        for i in range(len(left_end)):
            mask_ctx = tuple(j[left_end[i]:rigth_end[i]] for j in masks)
            context = "".join(rs.maskstoseq(mask_ctx))

            ctx.append(context)

        return ctx
    def context_generator(self, chromosome_object, k=1):
        """
        Return the context for a position in the mask. Needs the 3 masks
        in the correct order (nmask purine and strong). The k indicates
        how many positions we should move to the upstream and downstream.
        1 per default means the 96 context. It uses by default baseset
        C A, meaning that mutations that have a reference value diferent
        than C or A will be transformed to the other strand by reverese
        complementary
        """
        # Important pos_start != pos_end
        left_end = self.pos[:, 0] - k # this is the start
        rigth_end = self.pos[:, 1] + k  # this is the end

        masks = chromosome_object.seq_mask

        for i in range(len(left_end)):
            mask_ctx = tuple(j[left_end[i]:rigth_end[i]] for j in masks)
            context = "".join(rs.maskstoseq(mask_ctx))

            yield context

    def get_tr96(self, chromosome_object, k=1):
        " get the mutation subtype in MS96 format"
        ctx = self.get_context(chromosome_object, k=k)
        alt = self.get_alt()
        mut_list = []

        for i, val in enumerate(ctx):
            if val[k] in ["T", "G"]:
                ctx_tr = str(Seq(val).reverse_complement())
                alt_tr = str(Seq(alt[i]).reverse_complement())
            else:
                ctx_tr = val
                alt_tr = alt[i]

            mut = "{}>{}".format(ctx_tr, alt_tr)

            mut_list.append(mut)
        return mut_list
    def divide_batch(self, muts_x_batch):
        " Divide the current mutset object and yield different batches"
        total_muts = len(self)
        iterantions_n = total_muts // muts_x_batch
        iterantions_n = iterantions_n + 1
        
        first_positions = np.multiply(list(range(0, iterantions_n)),
                                      muts_x_batch)
        last_positions = np.multiply(list(range(1, iterantions_n + 1)),
                                     muts_x_batch)
        last_positions[-1] = total_muts
        # I think I do not have to touch this (substracting 1 ) because
        # I use the range functions that already does that.
        print("Number of iterations {}".format(iterantions_n))
        
        for i, s_idx in enumerate(first_positions):
            # s_idx = first_positions[i]
            e_idx = last_positions[i]
            #import pdb; pdb.set_trace()
            length_to_range = e_idx - s_idx
            strand_info = ["+" for i in range(length_to_range)]

            tmp = MutSet(chr_id=self.chr_id,
                         pos_start=self.pos[s_idx:e_idx, 0],
                         pos_end=self.pos[s_idx:e_idx, 1],
                         sample_id=self.meta[s_idx:e_idx, 0],
                         ref=self.meta[s_idx:e_idx, 1],
                         alt=self.meta[s_idx:e_idx, 2],
                         strand=strand_info,
                         to0base=False)

            yield tmp

def num2chr(chr_array):
    " Transform the 37 notation to hg19 "
    res = np.array(["chr{}".format(i) for i in chr_array])
    return res



"""
Functions to store and serialize refseqs
"""

import sys
import re
import numpy as np
from Bio import SeqIO



class RefSeq(object):
    """
    Class to store the original sequence of the Reference Sequence
    """
    def __init__(self, refseq_filepath, assembly):
        self.chr_refseq = SeqIO.to_dict(SeqIO.parse(refseq_filepath, "fasta"))
        self.assembly = assembly
    def __len__(self):
        return len(self.chr_refseq)
    def get_chr_names(self):
        " Return a list with the names of the chromosomes read "
        chrnames = [i for i in self.chr_refseq]
        return chrnames
    def dress_up_seq(self):
        " Mask the genome "
        n_re = re.compile(str("[N]"))
        purine_re = re.compile("[AaGg]")
        strong_re = re.compile("[CcGg]")
        chr_mask = {}

        sys.stderr.write("Starting Serialization\n")

        for chr_id in self.get_chr_names():

            sys.stderr.write("Starting Serialization of {}\n".format(chr_id))

            reference_seq = str(self.chr_refseq[chr_id].seq)

            n_mask = np.empty(len(reference_seq), dtype=bool)
            n_mask.fill(1) # all values to True
            idx = [i.start() for i in re.finditer(n_re, reference_seq)]
            n_mask[idx] = 0 # I put False when the position is masked

            purine_mask = np.empty(len(reference_seq), dtype=bool)
            purine_mask.fill(0) # all values to False
            idx = [i.start() for i in re.finditer(purine_re, reference_seq)]
            purine_mask[idx] = 1 # insert T when regex is matched

            strong_mask = np.empty(len(reference_seq), dtype=bool)
            strong_mask.fill(0) # all values to False
            idx = [i.start() for i in re.finditer(strong_re, reference_seq)]
            strong_mask[idx] = 1

            chr_mask[chr_id] = (n_mask, purine_mask, strong_mask)

        return chr_mask


def maskstoseq(masks):
    " Transform the 3 masks back to a sequence "
    res = np.empty(len(masks[0]), dtype=str)

    npos = np.logical_and(np.logical_not(masks[0]),
                          np.logical_and(np.logical_not(masks[1]),
                                         np.logical_not(masks[2])))

    gpos = np.logical_and(masks[0],
                          np.logical_and(masks[1],
                                         masks[2]))

    cpos = np.logical_and(masks[0],
                          np.logical_and(np.logical_not(masks[1]),
                                         masks[2]))

    apos = np.logical_and(masks[0],
                          np.logical_and(masks[1],
                                         np.logical_not(masks[2])))

    tpos = np.logical_and(masks[0],
                          np.logical_and(np.logical_not(masks[1]),
                                         np.logical_not(masks[2])))

    res[tpos] = "T"
    res[apos] = "A"
    res[cpos] = "C"
    res[gpos] = "G"

    res[npos] = "N"

    return res



"""
This module contains the main class for parsing the genome
This is a class that stores the 4 masks needed for the randomization process

At some point I can include some kind of automatic download from UCSC
"""

import randommut.refseq as rs


def genome_from_path(refseq_path, assembly):
    " File to load a genome directly from files "
    refseq_object = rs.RefSeq(refseq_path, assembly)
    refseq_mask = refseq_object.dress_up_seq()

    chr_list = [] # make this a dictionary?
    for i in refseq_object.get_chr_names():
        chr_tmp = Chromosome(seq_mask=refseq_mask[i], assembly=assembly, chr_id=i)
        chr_list.append(chr_tmp)
    return Genome(chr_list, assembly)

class Genome(object):
    """
    Class to store the 4 masks that will repreesent the available chromosome
    """
    def __init__(self, chr_list, assembly):
        self.chr_list = chr_list
        self.assembly = assembly

    def __len__(self):
        " Get the number of chromosomes stored in the genome "
        return len(self.chr_list)
    def chromosome_list(self):
        " Get the chromosome names "
        return [i for i in self.chr_list]
    def get_assembly(self):
        " Get the inputed assembly "
        return self.assembly
    def chromosome_iterator(self):
        " Yield a chromosome object in the list "
        for i in self.chr_list:
            yield i
    def get_chr_len(self):
        chr_len = {}
        for i in self.chromosome_iterator():
            chr_len[i.chr_id] = len(i)

        return chr_len

class Chromosome(object):
    """
    Class to store the 3 masks that will repreesent the available chromosome
    """
    def __init__(self, seq_mask, assembly, chr_id):
        self.assembly = assembly
        self.seq_mask = seq_mask
        self.chr_id = chr_id
    def n_mask(self):
        " returns the N mask, True is no N, False is N"
        return self.seq_mask[0]
    def purine_mask(self):
        " returns the purine mask, True is A G, False is C T"
        return self.seq_mask[1]
    def strong_mask(self):
        " returns the strong mask, True is C G, False is A T"
        return self.seq_mask[2]
    def __len__(self):
        "returns the length of the chromosome"
        return len(self.seq_mask[2])



"""
    RANDOM MUT
    ~~~~~~~~~~~~~

    This script uses the package randommut to randomize user set positions.

    :copyright: 2018 by my David Mas @ IRBBarcelona, AUTHORS for more details
    :license: license_name, see LICENSE for more details
"""

import sys
import argparse
import randommut.randommutworkflow as wf

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-M", "--mode", type=str,
                        choices=['serialize', 'randomize'],
                        help="The mode to run the software")
    parser.add_argument("-g", "--genome",
                        type=str,
                        help="The path to the reference sequence or serialized genome")
    parser.add_argument("-m", "--muts",
                        type=str,
                        help="The path to the muts file in avinput format",
                        default=None)
    parser.add_argument("-a", "--assembly",
                        type=str,
                        help="Assembly of the genome",
                        default="hg19")
    parser.add_argument("-o", "--outfile", default="rand_positions.tsv")
    parser.add_argument("-t", "--times", type=int, default=10,
                        help="Number of randomized positions")
    parser.add_argument("-w", "--winlen", type=int, default=50000,
                        help="Length of the windows to randomize")
    parser.add_argument("-b", "--batch_size", type=int, default=10000,
                        help="Length of the randomization batch. Decrease if memory error.")
    parser.add_argument("-C", "--compression", type=str,
                        choices=['gzip', 'bz2', None],
                        default=None,
                        help="If the output should be compressed")
    parser.add_argument("-v", "--verbose",
                        action="store_true",
                        default=False,
                        help="Script returns a more verbose messages.")

    args = parser.parse_args()

    if args.mode == "serialize":
        wf.serialize_genome(genome_path=args.genome, assembly=args.assembly)
    elif args.mode == "randomize":
        if args.muts is None:
            sys.exit('Mut file needed. add with -m tag')
        out_obj = wf.randomize(muts_path=args.muts,
                            genome_path=args.genome,
                            assembly=args.assembly,
                            times=args.times,
                            b_size=args.batch_size,
                            winlen=args.winlen,
                            verbose=args.verbose)

        wf.write_randomized_positions(randomize_output=out_obj,
                                    outfilename=args.outfile,
                                    compression=args.compression)
    else:
        sys.exit("Wrong mode inputed")


if __name__ == "__main__":
    main()



#!/usr/bin/env python

"""
    RANDOM MUT
    ~~~~~~~~~~~~~

    This script uses the package randommut to randomize user set positions.

    :copyright: 2018 by my David Mas @ IRBBarcelona, AUTHORS for more details
    :license: license_name, see LICENSE for more details
"""

import os
import sys
import pickle
import pandas as pd
from tqdm import trange
import randommut.genome as gn
import randommut.muts as mt
import randommut.randomize as rnd

def serialize_genome(genome_path, assembly):
    """
    Get a fasta file and transform to a pickle object
    """
    if genome_path.endswith(('.fa', '.fasta')):
        genome = gn.genome_from_path(genome_path, assembly)
        base = os.path.basename(genome_path)
        genome_path_pickle = "{}{}".format(base, ".p")

        pickle.dump(genome, open(genome_path_pickle, "wb"))
    else:
        raise ValueError

def randomize(muts_path, genome_path, assembly, times, winlen, verbose, b_size):
    """
    perform  the randomization
    """
    # Optimize this monster please
    # genome is a list of chromosomes (iteration through this)
    # mutset is a dictionary of chomosomes

    if genome_path.endswith('.p'):
        sys.stderr.write("Genome in pickle format detected\n")
        genome_path_pickle = genome_path
        genome = pickle.load(open(genome_path_pickle, "rb"))
        if genome.assembly != assembly:
            raise ValueError
    elif genome_path.endswith(('.fa', '.fasta')):
        genome = gn.genome_from_path(genome_path, assembly)

    sys.stderr.write("Genome read\n")

    muts = mt.mutset_from_path(muts_path)
    sys.stderr.write("Muts read\n")

    randomize_output = {}

    # adding the progress bar
    chromosome_list = genome.chr_list
    progress_chr = trange(len(chromosome_list))
    for chrom_idx in progress_chr:
        chrom = chromosome_list[chrom_idx]
        chr_id = chrom.chr_id
        progress_chr.set_description('Randomizing {}'.format(chr_id))
        if chr_id in muts:
            mutset = muts[chr_id]
            randomize_output[chr_id] = rnd.rand_single_chr_in_batch(
                chrom,
                mutset,
                times,
                winlen,
                batch_size=b_size,
                verbose=verbose)
        else:
            continue
    sys.stderr.write("Rand output generated\n")
    # recover all positions

    full_df = []
    for chrom in genome.chr_list: #opt
        chr_id = chrom.chr_id
        if chr_id in muts:
            mutset = muts[chr_id]
            pos_df = pd.DataFrame(mutset.pos)
            pos_df.columns = ['start', 'end']
            meta_df = pd.DataFrame(mutset.meta)
            meta_df.columns = ['sample', 'ref', 'alt']

            rand_out = randomize_output[chr_id]
            rand_df = pd.DataFrame(rand_out)
            rand_df.columns = ["R{}".format(i+1) for i in range(times)]

            tmp_full = pd.concat([pos_df, meta_df, rand_df], axis=1)
            #import pdb; pdb.set_trace()
            tmp_full["ctx"] = mutset.get_context(chrom) #opt
            tmp_full["chr"] = chr_id
            tmp_full["strand"] = 1
            # this should order the columns
            # chr start end strand ref alt R1 ... Rtimes
            cols = tmp_full.columns.tolist()
            cols = cols[-2:-1] + cols[0:2] + cols[-1:] + cols[2:3] + cols[3:5] + cols[-3:-2] + cols[5:-3]
            tmp_full = tmp_full[cols]
            full_df.append(tmp_full)
        else:
            continue

    # import pdb; pdb.set_trace()
    if len(full_df) > 1:
        final_df = full_df[0].append(full_df[1:])
    elif len(full_df) == 1:
        #can hapen if only 1 chromosome
        final_df = full_df[0]
    else:
        # this is the case described in issue #4
        raise ValueError('chromosomes in mutation file not present in reference file, check assemblies')
    return final_df

def write_randomized_positions(randomize_output, outfilename, compression):
    """
    write the df in a file using pandas
    """
    sys.stderr.write("Writting...\n")
    randomize_output.to_csv(outfilename,
                            sep="\t",
                            header=True,
                            index=False,
                            compression=compression)
    sys.stderr.write("Results file available at {}\n".format(outfilename))



from setuptools import setup

setup(name='randommut',
      version='0.1.3',
      description='A package to randomze somatic mutaitons in the genome',
      url='http://vm-ubuntu-dmp:3000/davidmasp/randommut',
      author='David Mas',
      author_email='david.mas@irbbarcelona.org',
      license='MIT',
      packages=['randommut'],
      install_requires=[
          'biopython','tqdm','numpy','pandas'
      ],
      zip_safe=False,
      entry_points = {
        'console_scripts': [
            'randommut = randommut.__main__:main'
        ]
       })



"""
Instructions to randomize the genome
"""

import sys
import numpy as np
from tqdm import tqdm
from Bio.Seq import Seq
#from memory_profiler import profile

def rand_single_chr_in_batch(chromosome_object,
                             mutset_object_all,
                             times,
                             winlen,
                             verbose,
                             batch_size=1000):
    """
    Calls the rand_single_chr in batch mode
    """
    if len(mutset_object_all) <= batch_size:
        rand_out = rand_single_chr(chromosome_object=chromosome_object,
                                   mutset_object=mutset_object_all,
                                   times=times,
                                   winlen=winlen,
                                   verbose=verbose)
    else:
        out_all = []
        for mutset_object in mutset_object_all.divide_batch(batch_size):
            out_tmp = rand_single_chr(chromosome_object=chromosome_object,
                                      mutset_object=mutset_object,
                                      times=times,
                                      winlen=winlen,
                                      verbose=verbose)
            out_all.append(out_tmp)
        rand_out = np.concatenate(out_all, axis=0)

    return rand_out



def rand_single_chr(chromosome_object, mutset_object, times, winlen, verbose):
    """
    Performs the rendomization process in just one chromosome
    """
    # 1. obtain matrix mask
    # 2. obtain contexts plus ctx_idx = {}
    # 3. determine operation per ctx
    # 4. loop though ctx_idx to obtain final matrix mask
    # 5. randomize
    # 6. get back to mutset idx

    #1
    mask_matrix_raw = generate_mask_matrix(mutset_object=mutset_object,
                                           chromosome_object=chromosome_object,
                                           winlen=winlen)

    #2
    contexts = mutset_object.get_context(chromosome_object)

    ctx_idx = {} # this is important to go back to the mutset ordering

    for idx, ctx in enumerate(contexts):
        ctx_revcomp = Seq(ctx).reverse_complement()
        ctx_revcomp = str(ctx_revcomp)
        if ctx in ctx_idx:
            ctx_idx[ctx].append(idx)
        elif ctx_revcomp in ctx_idx:
            ctx_idx[str(ctx_revcomp)].append(idx)
        else:
            ctx_idx[ctx] = [idx]

    #3 + 4
    ctx_matrix = {}
    for ctx in ctx_idx:
        if verbose:
            tqdm.write("{}\n".format(ctx))
        current_idx = ctx_idx[ctx]
        current_mask_matrix = [i[current_idx,] for i in mask_matrix_raw]

        # all this could go into a function FROM HERE
        # obtain the mask for ctx
        set_left = set(ctx[0])
        set_right = set(ctx[2])
        set_center = set(ctx[1])

        mask_left = compute_bimask(current_mask_matrix, set_left)
        mask_right = compute_bimask(current_mask_matrix, set_right)
        mask_center = compute_bimask(current_mask_matrix, set_center)

        mask_left_shifted = np.apply_along_axis(shift5,
                                                axis=1, # this should be rows
                                                arr=mask_left,
                                                num=1, # this moves to the right
                                                fill_value=False)

        mask_right_shifted = np.apply_along_axis(shift5,
                                                 axis=1, # this should be rows
                                                 arr=mask_right,
                                                 num=-1, # this moves to  left
                                                 fill_value=False)

        mask_ends_context = np.logical_and(mask_left_shifted,
                                           mask_right_shifted)
        mask_context_sense = np.logical_and(mask_center, mask_ends_context)
        # TO HERE

        # obtain the mask for reverse complement
        ctx_revcomp = Seq(ctx).reverse_complement()

        set_left = set(ctx_revcomp[0])
        set_right = set(ctx_revcomp[2])
        set_center = set(ctx_revcomp[1])

        mask_left = compute_bimask(current_mask_matrix, set_left)
        mask_right = compute_bimask(current_mask_matrix, set_right)
        mask_center = compute_bimask(current_mask_matrix, set_center)

        mask_left_shifted = np.apply_along_axis(shift5,
                                                axis=1, # this should be rows
                                                arr=mask_left,
                                                num=1, # this moves to the right
                                                fill_value=False)

        mask_right_shifted = np.apply_along_axis(shift5,
                                                 axis=1, # this should be rows
                                                 arr=mask_right,
                                                 num=-1, # this moves to  left
                                                 fill_value=False)

        mask_ends_context = np.logical_and(mask_left_shifted,
                                           mask_right_shifted)

        mask_context_reverse = np.logical_and(mask_center, mask_ends_context)

        mask_final = np.logical_or(mask_context_reverse, mask_context_sense)

        number_ctx = np.sum(mask_final)
        if verbose:
            tqdm.write("Number of muts in context: {}\n".format(number_ctx))
        ctx_matrix[ctx] = mask_final

    # 5
    rand_res = {}
    for ctx in ctx_matrix:
        rand_result = randomize_mask_matrix(ctx_matrix[ctx], times=times)
        rand_res[ctx] = rand_result

    # 6
    pos = mutset_object.pos
    left_end = pos[:, 0] - winlen # start minus width

    # this is the output, then the cols are max 100 (no memory issue here)
    rand_pos = np.empty((len(pos), times), dtype=int)
    rand_pos.fill(0) # safety first

    for ctx in rand_res:
        idx = ctx_idx[ctx]
        # we go back to oiginal
        original_positions = np.array(left_end[idx]).reshape([len(idx), 1])
        rand_pos[idx, ] = rand_res[ctx] + original_positions
    return rand_pos

# extracted from https://stackoverflow.com/a/42642326/5410410
# preallocate empty array and assign slice by chrisaycock
def shift5(arr, num, fill_value=np.nan):
    """
    function that shifts num elements of a numpy array to the right (positive)
    or left (negative) and incorporates in the missing spots the fill-value
    argument (nan) by default.
    """
    result = np.empty_like(arr)
    if num > 0:
        result[:num] = fill_value
        result[num:] = arr[:-num]
    elif num < 0:
        result[num:] = fill_value
        result[:num] = arr[-num:]
    else:
        result = arr
    return result


def mask_to_pvector(mask):
    """
    It transforms a mask object (np.boolean array) into a probability vector of
    the same size with an equal value for all the unmasked indexs equivalent
    to the 1 / (number of masked elements). Thus,
    T T T F F F T F F F
    will be
    0.25 0.25 0.25 0 0 0 0.25 0 0 0
    """
    pvector = np.zeros(len(mask)) # the zeros function add 0 as a float

    total_pos = np.sum(mask)

    if total_pos != 0:
        prob_val = 1 / total_pos
        pvector[mask] = prob_val
    elif total_pos == 0:
        pvector.fill(-1)
    else:
        raise ValueError

    # necessary? it makes the function crash, choice already handles this
    #if pvector.sum() != 1:
    #    raise ValueError

    return pvector


def randomize_mask_row(mask, times):
    """
    Here we are tranforming a definitive mask into a pvector and then
    chosing randomly the idx that will be selected. These are equivalent
    to the relatives positions in the sequence.
    """
    # why replace = True? The idea is that different randomizations are pure
    # independent experiments, then, this should be rep true.
    pvector = mask_to_pvector(mask)
    int_size = len(pvector) # this should be equibalent to the wl*2

    if pvector[0] != -1:
        rand_idx = np.random.choice(int_size,
                                    p=pvector,
                                    size=times,
                                    replace=True)
    else:
        rand_idx = np.zeros(times)

    return rand_idx

def randomize_mask_matrix(mask_matrix, times):
    """
    It applies the function randomize_mask_row row by row in a
    definitive Mutset Mask
    """
    idx_matrix = np.apply_along_axis(randomize_mask_row,
                                     axis=1, # this is rows
                                     arr=mask_matrix,
                                     times=times) # using the kargs argument

    return idx_matrix

def compute_bimask(masks, biset):
    """
    Internally transforms inputed masks to desidred bisets of bases. This
    mean that I can transform the 3 masks in a mask that contains all G and Ts.
    """
    # masks reminder
    # index 0 - n mask
    # index 1 - purine (AG)
    # index 2 - strong (CG)

    strong_set = set(["C", "G"])
    weak_set = set(["A", "T"])

    purine_set = set(["A", "G"])
    pyrimidine_set = set(["C", "T"])

    gt_set = set(["G", "T"])
    ca_set = set(["C", "A"])

    c_set = set(["C"])
    a_set = set(["A"])
    t_set = set(["T"])
    g_set = set(["G"])

    n_set = set(["N"])
    # I check which option are we talking about and do the appropiate masks
    if biset == ca_set:
        mask_res = np.logical_xor(masks[1], masks[2])
    elif biset == gt_set:
        mask_res = np.logical_xor(masks[1], masks[2])
        mask_res = np.logical_not(mask_res)
    elif biset == strong_set:
        mask_res = masks[2]
    elif biset == weak_set:
        mask_res = np.logical_not(masks[2])
    elif biset == pyrimidine_set:
        mask_res = np.logical_not(masks[1])
    elif biset == purine_set:
        mask_res = masks[1]
    elif biset == c_set:
        not_purine = np.logical_not(masks[1])
        mask_res = np.logical_and(not_purine, masks[2])
    elif biset == t_set:
        not_purine = np.logical_not(masks[1])
        not_strong = np.logical_not(masks[2])
        mask_res = np.logical_and(not_purine, not_strong)
    elif biset == g_set:
        mask_res = np.logical_and(masks[2], masks[1])
    elif biset == a_set:
        not_strong = np.logical_not(masks[2])
        mask_res = np.logical_and(masks[1], not_strong)
    elif biset == n_set:
        # see isssue #1 in gitea
        # I am setting the whole mask to F
        # then I add a step in the rand process when if this happens
        # I generate -1 as randomized positions.
        # update: see n. asjasdhajs
        mask_res = np.zeros(np.shape(masks[0]), dtype=bool)
    else:
        sys.stderr.write(biset)
        raise ValueError

    # important to not forget the n_mask
    mask_final = np.logical_and(mask_res, masks[0])

    return mask_final

#@profile
def generate_mask_matrix(mutset_object, chromosome_object, winlen):
    """
    It generates mask matrix
    """
    winlen = winlen + 1 # I think (due to the 0 based stuff)
    total_length = winlen + winlen + 1 # this because of the middle position

    # here is the memory issue
    # NOTE asjasdhajs
    # trying to solve it with this
    # np.ones((2, 2), dtype=bool)
    # from here
    # https://stackoverflow.com/a/21174962/5410410
    # After some test in this approach we reduce the virtual memory generation
    # in the initialitzation phase (reduction of half)
    # Though the RES memory is the same and it behaves similar, with a
    # reduction after a while. I guess this is numpy storing a number vs
    # when it realises in only a boolean array. So it is an improvement I g.
    # Not final solution though.

    n_muts = len(mutset_object.pos)

    mask_matrix = [np.zeros([n_muts, total_length], dtype=bool),
                   np.zeros([n_muts, total_length], dtype=bool),
                   np.zeros([n_muts, total_length], dtype=bool)]

    masks = chromosome_object.seq_mask

    chromosome_length = len(masks[0])

    for j, original_mask in enumerate(masks):
        for i, val in enumerate(mutset_object.pos): #this should enumerate rows
            left_end = val[0] - winlen # this is the start and then i substract
            right_end = val[1] + winlen # this is the end and then i add

            if left_end < 0:
                #tqdm.write("SHORT position found \n")
                false_pos = winlen - val[0]
                # see update in asjasdhajs
                #mask_matrix[j][i, :false_pos] = False
                mask_matrix[j][i, false_pos:] = original_mask[:right_end]
            elif right_end > chromosome_length:
                #tqdm.write("LONG position found \n")
                false_pos = int(right_end) - int(chromosome_length) #this is p
                end_of_world = total_length - false_pos
                mask_matrix[j][i, :end_of_world] = original_mask[left_end:chromosome_length]
                #mask_matrix[j][i, end_of_world:] = False
                # see update in  asjasdhajs
            else:
                mask_matrix[j][i,] = original_mask[left_end:right_end]

    return mask_matrix
